
struct BckKey
(
  time, -- float 
  value, -- float 
  tangent -- float  //??
)

struct BckJointAnim
(
  scalesX = #(), -- std::vector<Key>
  scalesY = #(), -- std::vector<Key>
  scalesZ = #(), -- std::vector<Key>

  rotationsX = #(), -- std::vector<Key>
  rotationsY = #(), -- std::vector<Key>
  rotationsZ = #(), -- std::vector<Key>

  translationsX = #(), -- std::vector<Key>
  translationsY = #(), -- std::vector<Key>
  translationsZ = #() -- std::vector<Key>
)


------------------------------------

struct BckAnk1Header
(
  tag, -- char[4] 'ANK1'
  sizeOfSection, -- u32 

  -- 0 - play once, 2 - loop
  loopFlags, -- u8 

  angleMultiplier, -- u8 all angles have to multiplied by pow(2, angleMultiplyer)

  animationLength, -- u16 in time units

  numJoints, -- u16 that many animated joints at offsetToJoints
  scaleCount, --u16  that many floats at offsetToScales
  rotCount, -- u16 that many s16s at offsetToRots
  transCount, -- u16 that many floats at offsetToTrans

  offsetToJoints, -- u32 
  offsetToScales, -- u32 
  offsetToRots, -- u32 
  offsetToTrans, -- u32 
  
  fn LoadData br = 
  (
  	  tag = br.ReadFixedLengthString 4
  	  sizeOfSection = br.ReadDWORD()
	  loopFlags = br.GetByte()
	  angleMultiplier = br.GetByte()
	  animationLength = br.ReadWORD()
	  numJoints = br.ReadWORD()
	  scaleCount = br.ReadWORD()
	  rotCount = br.ReadWORD()
	  transCount = br.ReadWORD()
	  offsetToJoints = br.ReadDWORD()
	  offsetToScales = br.ReadDWORD()
	  offsetToRots = br.ReadDWORD()
	  offsetToTrans = br.ReadDWORD()
	
  )
)


-- TODO: the following two structs have really silly names, rename them
struct BckAnimIndex
(
  count, -- u16 
  index, -- u16 
  zero, -- u16 always 0?? -> no (biawatermill01.bck) TODO: find out what it means
  
  fn LoadData br =
  (
     count = br.GetSHORT()
	 index = br.GetSHORT()
	 zero = br.GetSHORT()
  )
)

struct BckAnimComponent
(
  s = BckAnimIndex(), -- AnimIndex scale
  r = BckAnimIndex(), -- AnimIndex rotation
  t = BckAnimIndex(), -- AnimIndex translation
  
  fn LoadData br =
  (
     s.LoadData br
	 r.LoadData br
	 t.LoadData br
  )
)

struct BckAnimatedJoint
(

  --if count > 1, count*3 floats/shorts stored at index (time, value, unk [interpolation info, e.g. tangent??])?
  --for shorts, time is a "real" short, no fixedpoint
  x = BckAnimComponent(), -- AnimComponent 
  y = BckAnimComponent(), -- AnimComponent 
  z = BckAnimComponent(), -- AnimComponent 
  
  fn LoadData br =
  (
	 x.LoadData br
	 y.LoadData br
	 z.LoadData br
  )
)

-----------------------------------------------

struct Bck
(
  anims = #(), -- std::vector<JointAnim>
  animationLength, -- int 
  currAnimTime, -- float 
  
  -- ConvRotation(vector<Key>& rots, float scale)
  fn ConvRotation rots scale =
  (
     for j = 1 to rots.count do
	 (
	     rots[j].value *= scale
		 rots[j].tangent *= scale
	 )
	 
	 return rots
  ),
  
  -- void readComp(vector<Key>& dst, const vector<T>& src, const bck::AnimIndex& index)
  fn ReadComp dst src index =
  (
     dst = #() -- dst.resize(index.count);
	 
	  -- violated by biawatermill01.bck
	  if index.zero != 0 then
	  (
	    throw "bck: zero field %d instead of zero"
	    --TODO: biawatermill01.bck doesn't work, so the "zero"
	    --value is obviously something important
	  )
	  
	  if index.count <= 0 then
	  (
	    print "Warning: readComp(): count is <= 0"
	  )
	  else if index.count == 1 then
	  (
	    dst[1] = BckKey()
	    dst[1].time = 0
	    dst[1].value = src[index.index + 1]
	    dst[1].tangent = 0
	  )
	  else
	  (
	    for j = 0 to (index.count - 1) do -- (int j = 0; j < index.count; ++j)
	    (
		  dst[j + 1] = BckKey()
	      dst[j + 1].time = src[(index.index + 3*j) + 1]
	      dst[j + 1].value = src[(index.index + 3*j + 1) + 1]
	      dst[j + 1].tangent = src[(index.index + 3*j + 2) + 1]
	    )
	  )
  
  	return dst
  ),
  
  fn LoadAnk1 br =
  (

      i = 0
	  ank1Offset = br.Position()
	  
	  -- read header
      h = BckAnk1Header()
	  h.LoadData br
	  currAnimTime = 0.0
	  animationLength = h.animationLength
	  
	  -- read scale floats:
	  br.SeekSet (ank1Offset + h.offsetToScales)
	  scales = #() -- vector<f32> scales(h.scaleCount);
	  for i=1 to h.scaleCount do
	      append scales (br.GetFloat())

	  -- read rotation s16s:
	  br.SeekSet (ank1Offset + h.offsetToRots)
	  rotations = #()
	  for i=1 to h.rotCount do
	       append rotations (br.GetSHORT())
		   
	  -- read translation floats:
	  br.SeekSet (ank1Offset + h.offsetToTrans)
	  translations= #() 
	  for i=1 to h.transCount do
	      append translations (br.GetFloat()) 

	  -- read joints
	  rotScale = (pow (2 as float) h.angleMultiplier) * 180 / (32768 as float)
	  br.SeekSet (ank1Offset + h.offsetToJoints)
	  
	  anims = #() -- bck.anims.resize(h.numJoints);
	  for i = 1 to h.numJoints do
	  (
	    joint = BckAnimatedJoint()
		joint.LoadData br 
		
		anims[i] = BckJointAnim()

	    anims[i].scalesX = ReadComp anims[i].scalesX scales joint.x.s
	    anims[i].scalesY = ReadComp anims[i].scalesY scales joint.y.s
	    anims[i].scalesZ = ReadComp anims[i].scalesZ scales joint.z.s
	
	    anims[i].rotationsX = ReadComp anims[i].rotationsX rotations joint.x.r
	    anims[i].rotationsX = ConvRotation anims[i].rotationsX rotScale
	    anims[i].rotationsY = ReadComp anims[i].rotationsY rotations joint.y.r
	    anims[i].rotationsY = ConvRotation anims[i].rotationsY rotScale
	    anims[i].rotationsZ = ReadComp anims[i].rotationsZ rotations joint.z.r
	    anims[i].rotationsZ = ConvRotation anims[i].rotationsZ rotScale
	
	    anims[i].translationsX = ReadComp anims[i].translationsX translations joint.x.t
	    anims[i].translationsY = ReadComp anims[i].translationsY translations joint.y.t
	    anims[i].translationsZ = ReadComp anims[i].translationsZ translations joint.z.t
	  )
	  
  ),
  
  fn LoadBck filePath =
  (
      local br = BinaryReader()
	  br.Open filePath 
	  br.SeekSet 0x20 
	  local pos = 0
	  local size = 0
	  local i = 0
	  do
	  (
		br.SeekCur size
	    local pos = br.Position()
	    local tag = br.ReadFixedLengthString 4
		size = br.ReadDWORD()

		br.SeekSet pos

	    if tag == "ANK1" then
	      LoadAnk1 br
	    else
	      throw ("readBck(): Unsupported section " + tag)
	
	    br.SeekSet pos

		i += 1
	  ) while i < 1
	  
  
  ),
  
  
  fn Interpolate v1 d1 v2 d2 t = -- t in [0,1]
  (
  	   --cubic interpolation
	   -- float values
	  a = 2*(v1 - v2) + d1 + d2
	  b = -3*v1 + 3*v2 - 2*d1 - d2
	  c = d1
	  d = v1
	  --TODO: yoshi_walk.bck has strange-looking legs...not sure if
  	 --the following line is to blame, though
  	 return ((a*t + b)*t + c)*t + d
  ),
  
  fn GetAnimValue keys t =
  (
	  if keys.count == 0 then
	    return 0.0
	
	  if keys.count == 1 then
	    return keys[1].value
	
	  i = 2
	  while keys[i].time < t do
	    i += 1
	
	  time = (t - keys[i - 1].time)/(keys[i].time - keys[i - 1].time) -- scale to [0, 1]
	  return interpolate keys[i - 1].value keys[i - 1].tangent keys[i].value keys[i].tangent time
  ),
  
  
  



-- the caller has to ensure that jnt1.frames and bck.anims contain
--the same number of elements
fn AnimateJnt jnt deltaTime =
(
  -- update time
  currAnimTime += deltaTime*16; -- convert from seconds to ticks (dunno if this is right this way...TODO)
  currAnimTime = mod currAnimTime animationLength -- loop?

  -- update joints
  for i = 1 to jnt.frames.count do
  (
    jnt.frames[i].sx = getAnimValue anims[i].scalesX currAnimTime
    jnt.frames[i].sy = getAnimValue anims[i].scalesY currAnimTime
    jnt.frames[i].sz = getAnimValue anims[i].scalesZ currAnimTime

    --TODO: use quaternion interpolation for rotations?
    jnt.frames[i].rx = getAnimValue anims[i].rotationsX currAnimTime
    jnt.frames[i].ry = getAnimValue anims[i].rotationsY currAnimTime
    jnt.frames[i].rz = getAnimValue anims[i].rotationsZ currAnimTime

    jnt.frames[i].t.x = getAnimValue anims[i].translationsX currAnimTime
    jnt.frames[i].t.y = getAnimValue anims[i].translationsY currAnimTime
    jnt.frames[i].t.z = getAnimValue anims[i].translationsZ currAnimTime
  )
),

fn AnimateBoneFrames bones frameScale =
(
	animate on
	(
	  for i = 1 to bones.count do
	  (
	  
	  	bone = bones[i]
		anim = anims[i]
		
		for j = 1 to anim.rotationsX.count do
		(
			rot = anim.rotationsX[j]
			at time (rot.time * frameScale) (bone.rotation.controller.x_rotation = rot.value )
		)
		
		for j = 1 to anim.rotationsY.count do
		(
			rot = anim.rotationsY[j]
			at time (rot.time * frameScale) (bone.rotation.controller.y_rotation = rot.value )
		)
		
		for j = 1 to anim.rotationsZ.count do
		(
			rot = anim.rotationsZ[j]
			at time (rot.time * frameScale) (bone.rotation.controller.Z_rotation = rot.value )
		)
		
		for j = 1 to anim.translationsX.count do
		(
			t  = anim.translationsX[j]
			at time (t.time * frameScale) (bone.position.controller.x_position = t.value )
		)
		
		for j = 1 to anim.translationsY.count do
		(
			t  = anim.translationsY[j]
			at time (t.time * frameScale) (bone.position.controller.y_position = t.value )
		)
		
		for j = 1 to anim.translationsZ.count do
		(
			t  = anim.translationsZ[j]
			at time (t.time * frameScale) (bone.position.controller.z_position = t.value )
		)
		
		
	  )
	)

	animationRange = interval 0 (animationLength * frameScale)
),


fn AnimateBones bones deltaTime =
(
  -- update time
  currAnimTime += deltaTime*16; -- convert from seconds to ticks (dunno if this is right this way...TODO)
  currAnimTime = mod currAnimTime animationLength -- loop?

  -- update joints
  for i = 1 to bones.count do
  (
  
  	bone = bones[i]
	
    sx = getAnimValue anims[i].scalesX currAnimTime
    sy = getAnimValue anims[i].scalesY currAnimTime
    sz = getAnimValue anims[i].scalesZ currAnimTime

    --TODO: use quaternion interpolation for rotations?
    rx = getAnimValue anims[i].rotationsX currAnimTime
    ry = getAnimValue anims[i].rotationsY currAnimTime
    rz = getAnimValue anims[i].rotationsZ currAnimTime

	bone.rotation.controller.x_rotation = rx
	bone.rotation.controller.y_rotation = ry
	bone.rotation.controller.z_rotation = rz
	
	
    tx = getAnimValue anims[i].translationsX currAnimTime
    ty = getAnimValue anims[i].translationsY currAnimTime
    tz = getAnimValue anims[i].translationsZ currAnimTime
	
	bone.position.controller.x_position = tx
	bone.position.controller.y_position = ty
	bone.position.controller.z_position = tz
	
	
  )
),

-- from stdplugs/stdscripts/CharacterPluginObject.ms
fn resetAnim fromAnim =
(
	if fromAnim.controller != undefined then
		deleteKeys fromAnim.controller #allKeys			
	
	-- delete anims from custom attributes
	for k=1 to (custAttributes.count fromAnim) do 
	(
		local ca = custattributes.get fromAnim k 
		if (ca != undefined) do
		(
			local saNames = getSubAnimNames ca

			for s=1 to saNames.count do 
			(
				if (ca[s].controller != undefined) do (
					deleteKeys ca[s].controller #allKeys
				)	
			)
		)
	)
	
	for j=1 to fromAnim.numSubs do
	(
		resetAnim fromAnim[j]
	)
),

-- from stdplugs/stdscripts/CharacterPluginObject.ms
fn resetNodeAnim node = 
(
	resetAnim node.controller
	resetAnim node.baseObject
	for m in node.modifiers do (
		resetAnim m
	)
)

)
