
struct FrameNode
(
    name,
	startPoint, -- end point = first child
	children = #(), -- FrameNode
	 parentFrameNode,
	 effP,
	f,
	 _bone,
	_eulerController, -- used for animations

	-- private
	 fn _PrintTree depth =
	(
	   if (name != undefined) then
	   		print (depth + name)
		
		for child in children do
		   child._PrintTree (depth + "--")
	),
	
	 -- used for testing
	 fn PrintTree =
	(
		_PrintTree ""
	),
	
	-- private
	fn _CreateBones parentBone boneThickness createdBonesTable postfixName parentTransform =
	(
	     local bone = undefined
		 local endPoint = undefined
		 
	     if (parentFrameNode != undefined) then
		 (
			   if (children.count > 0 ) then
			 (
	      		endPoint = children[1].startPoint
			 )
			 
			 else
			(
				-- create an end point bone (same direction as parent bone with a length of boneThickness)
			    local start = startPoint 
				dir  = normalize (parentBone.position - start) 
				dir  *= (-1 * boneThickness)
				dir += startPoint 
				endPoint = [dir.x, dir.y, dir.z]
			)
		   		
		   if (parentBone == undefined) then
			(
				endPoint = [startPoint.x, startPoint.y, startPoint.z + boneThickness] 
			)
			
			--startPoint = [0,0,0]
			--endPoint = [10, 0, 0]
			
			 bone = bonesys.createbone startPoint endPoint z_axis
			_bone = bone

			_bone.scale.controller = ScaleXYZ ()
			
			
			mTransform = matrix3 1 -- identity
			mt = transMatrix [f.t.x, f.t.y, f.t.z]
			mx = rotateXMatrix f.rx
			my = rotateYMatrix f.ry
			mz = rotateZMatrix f.rz
			
			if (parentBone != undefined) then
			(
				mTransform =  (mx * my * mz * mt) * parentTransform
				
				bone.transform = mTransform
				bone.parent = parentBone
			)
			else
			(
				mTransform =  (mx * my * mz * mt) 
				bone.transform = mTransform
			)
			
			  bone.name = name + postfixName 
			  bone.width = boneThickness
			  bone.height = boneThickness
			  append createdBonesTable bone
	 
		 )		 
		 
		 for child in children do
		   child._CreateBones bone boneThickness createdBonesTable postfixName  mTransform
	),
	
	fn _ApplyRotation zDir =
	(
			if (_bone != undefined ) then
			(
				
			
				
				if (_bone.parent != undefined ) then 
				(
					m = Euler_XYZ()
					m.x_rotation = f.rx
					m.y_rotation = f.ry
					m.z_rotation = f.rz
					
					_bone.scale.controller = ScaleXYZ () -- resets animations
					
					_eulerController = m
					_bone.rotation.controller = m 	

					pos = Position_XYZ()
					pos.x_position = f.t.x 
					pos.y_position = f.t.y
					pos.z_position = f.t.z
					
					_bone.position.controller = pos
				)
			)
			
		 for child in children do
		   child.ApplyRotation()
	),
	
	
	fn ApplyRotation =
	(_ApplyRotation 1),
	
	-- private: NYI
	fn _FixBoneLength =
	(
		for childBone in children do
		(
		
			if (_bone != undefined AND _bone.parent != undefined) then
			(
				mt = transMatrix [f.t.x, f.t.y, f.t.z]
				mx = rotateXMatrix f.rx
				my = rotateYMatrix f.ry
				mz = rotateZMatrix f.rz
				
				parentVec = normalize ((in coordsys world _bone.position) - (in coordsys world _bone.parent.position))
			
				mTransform =  (mx * my * mz * mt) 
				
				boneVec2 = normalize ((normalize _bone.dir) * mTransform)
				
				print (((parentVec as string) + ":" + (boneVec2 as string)))
			)
			
			childBone._FixBoneLength()
		)
	),
	
	-- returns new bones array
	fn CreateBones boneThickness postfixName =
	(
		if postfixName == undefined then
			postfixName = ""
	   createdBonesTable = #()
		
		mTransform = matrix3 1 -- identity
			
	    _CreateBones undefined boneThickness createdBonesTable postfixName mTransform
		return createdBonesTable 
	),
	
	fn _RemapBones boneSet =
	(
		_bone = getNodeByName name
		append boneSet _bone 
		for child in children do
		   child._RemapBones boneSet
	),
	
	-- used on hold, fetch. Bone references lost.
	fn RemapBones =
	(
		local boneSet = #()
		 for child in children do
		   child._RemapBones boneSet
		 
		 return boneSet
	)
)
