

struct Mat3Header
(
   tag, -- char[4] 'MAT3'
  sizeOfSection, -- u32 
   count, -- u16
   pad, -- u16

  /*
    0 - MatInit array
    1 - array of shorts with 0, 1, 2, ..., count (nearly...sometimes small deviations) -> index into offset[0]
    2 - string table
    3 - divisible by Mat3Header.count - so the Mat3Entries are stored here
    15 - index to texture table?
  */
  offsets = #(), -- u32 [30]
	
	fn LoadData br =
	(
		tag = br.ReadFixedLengthString 4
		sizeOfSection = br.ReadDWORD()
		count = br.ReadWORD()
		pad = br.ReadWORD()
		
		for i = 1 to 30 do
			offsets[i] = br.ReadDWORD()
	)
)

struct MatInit
(
   unknown1 = #(), -- u8 [132]
   texStages = #(), -- u16[8]
  unknown2 = #(), -- u8 [332 - 132 - 8*2]
	
	fn LoadData br =
	(
		for i = 1 to 132 do
			unknown1[i] = br.GetByte()
		
		for i = 1 to 8 do
			texStages[i] = br.ReadWORD()
		
		u2Count = 332 - 132 - 8*2
		for i = 1 to u2Count do
			unknown2[i] = br.GetByte()
	)
)



struct MatEntry
(
  -- size = 312 = 0x138
  data -- u8 data[312], -- 
)



struct Mat3
(
  -- temporary, maps mat index to tex index
  texTable = #(), -- std::vector<int> 
	stringtable,
	
	fn LoadData br =
	(
		-- "Mat3 section support is very incomplete"
		
		mat3Offset = br.Position()

	  -- read header
		h = Mat3Header()
		h.LoadData br

		stringtable = br.ReadStringTable (mat3Offset + h.offsets[3]) -- readStringtable(mat3Offset + h.offsets[2], f, stringtable);

		if h.count != stringtable.count then
			throw "mat3: number of strings (%d) doesn't match number of elements (%d)"

		  -- compute max length of each subsection
		  lengths = #() -- vector<int> lengths(30);
		  for i = 1 to 30 do
		  (
				next = h.sizeOfSection
				
				if h.offsets[i] != 0 then
				(
					  for j = i + 1 to 30 do
					  (
							if h.offsets[j] != 0 then
							(
							  next = h.offsets[j]
							  break
							)
					  )
				)
				else
				(
				  next = 0
				)
				
				len = next - h.offsets[i]
				lengths[i] = len
				if (i == 3) then
				(
				  -- assert(length%h.count == 0); //violated by luigi's mansion files
				  -- assert(length/h.count == 312); //violated by quite a few files
				)
		  )
		  
		  
		  br.SeekSet (mat3Offset + h.offsets[2]) -- offset[1] (indirection table from indices to init data indices)
		  maxIndex = 0
		  indexToInitData = #() -- vector<int> indexToInitData(h.count);
		  for i = 1 to h.count do
		  (
				bla = br.ReadWORD()
				if (bla > maxIndex) then
					maxIndex = bla
				indexToInitData[i] = bla
		  )

		  br.SeekSet (mat3Offset + h.offsets[1])

		  initData = #() -- vector<bmd::MatInit> initData(maxIndex + 1)
		  
		  for i = 1 to (maxIndex + 1) do -- for(i = 0; i <= maxIndex; ++i)
		  (
				init = MatInit()
				init.LoadData br
			  initData[i] = init
		  )

		  -- read texTable
		br.SeekSet (mat3Offset + h.offsets[16]) --  fseek(f, mat3Offset + h.offsets[15], SEEK_SET);
		 
		   texLength = lengths[16]
		  texTable = #() -- vector<int> texTable(texLength/2);
		   maxTexIndex = 0;
			  
		  for i = 1 to (texLength/2) do
		  (
				index = br.ReadWORD()
			  texTable[i] = index
			  
			  if (index > maxTexIndex) then
				maxTexIndex = index
		  )

		  for i = 1 to h.count do
		  (
				stage =  initData[indexToInitData[i] + 1].texStages[1] 
				if (stage != 0xffff) then
				  texTable[i] = texTable[stage + 1]
				else
				  texTable[i] = 0xffff
		  )
  
	)
)